use rocket::{
	form::{self, Contextual, Form},
	response::Redirect,
};
use rocket_dyn_templates::{context, Template};
use uuid::uuid;

use crate::{
	db::{
		doll,
		schema::{CreateDollProfile, DollTagsDb},
	},
	ids::{id_public_to_db, pick_ids},
	pages::CommonTemplateState,
	routes::{error_handlers::PageResult, public},
};

#[get("/register_tag")]
pub async fn show_register(db: DollTagsDb, meta: CommonTemplateState) -> PageResult {
	let ids = pick_ids(db).await?;

	Ok(Template::render(
		"register_tag",
		context! {
			ids,
			previous: form::Context::default(),
			meta,
		},
	)
	.into())
}

#[derive(Debug, FromForm)]
pub struct TagForm<'a> {
	#[field(validate=validate_id())]
	pub ident: &'a str,
	#[field(validate=len(..32))]
	pub microchip_id: &'a str,

	#[field(validate=len(1..=256))]
	pub name: &'a str,
	#[field(validate=len(1..=32))]
	pub pronoun_subject: &'a str,
	#[field(validate=len(1..=32))]
	pub pronoun_object: &'a str,
	#[field(validate=len(1..=32))]
	pub pronoun_possessive: &'a str,

	#[field(validate=len(1..=256))]
	pub handler_name: &'a str,
	#[field(validate=len(..=2048))]
	pub handler_link: &'a str,

	#[field(validate=len(..=256))]
	pub kind: &'a str,
	#[field(validate=len(..=256))]
	pub breed: &'a str,
	#[field(validate=len(..=256))]
	pub behaviour: &'a str,
	#[field(validate=len(..=2048))]
	pub description: &'a str,

	#[field(validate=validate_chassis(self.chassis_id, self.chassis_color, "chassis_type"))]
	pub chassis_type: &'a str,
	#[field(validate=validate_chassis(self.chassis_type, self.chassis_color, "chassis_id"))]
	pub chassis_id: &'a str,
	#[field(validate=validate_chassis(self.chassis_type, self.chassis_id, "chassis_color"))]
	pub chassis_color: &'a str,
}

fn validate_id<'v>(id: &str) -> form::Result<'v, ()> {
	if let None = id_public_to_db(id) {
		Err(form::Error::validation("id not formatted properly"))?;
	}

	Ok(())
}

fn validate_chassis<'v>(a: &str, b: &str, c: &str, field: &str) -> form::Result<'v, ()> {
	let all_empty = a.len() == 0 && b.len() == 0 && c.len() == 0;
	let all_full = a.len() != 0
		&& a.len() <= 256
		&& b.len() != 0
		&& b.len() <= 256
		&& c.len() != 0
		&& c.len() <= 256;

	if !all_empty && !all_full {
		Err(form::Error::validation(format!("missing {}", field)))?;
	}

	Ok(())
}

#[post("/register_tag", data = "<tag>")]
pub async fn handle_register(
	mut db: DollTagsDb,
	tag: Form<Contextual<'_, TagForm<'_>>>,
	meta: CommonTemplateState,
) -> PageResult {
	let tag = match tag.value {
		Some(ref values) => values,
		None => {
			// in case the form validation fails, this will be tasked with rendering the page again with submitted values and display errors
			let ids = pick_ids(db).await?;

			debug!("tag registration form invalid, context: {:?}", &tag.context);

			return Ok(Template::render(
				"register_tag",
				context! {
					ids,
					previous: &tag.context,
					meta,
				},
			)
			.into());
		}
	};

	debug!("registering tag: {:?}", tag);
	fn normalize_opt(opt: &str) -> Option<&str> {
		if opt.len() != 0 {
			Some(opt)
		} else {
			None
		}
	}

	let id = id_public_to_db(&tag.ident).expect("id format was wrong but is now right??");
	let normalized_microchip_id = tag.microchip_id.to_lowercase();
	let microchip_id = normalize_opt(&normalized_microchip_id);

	if doll::id_exists(&mut *db, id, microchip_id.unwrap_or("")).await? {
		return Ok(Redirect::found(uri!(show_register)).into());
	}

	doll::create(
		db,
		CreateDollProfile {
			id,
			microchip_id,
			name: tag.name,
			pronoun_subject: tag.pronoun_subject,
			pronoun_object: tag.pronoun_object,
			pronoun_possessive: tag.pronoun_possessive,
			handler_name: tag.handler_name,
			handler_link: normalize_opt(tag.handler_link),
			kind: normalize_opt(tag.kind),
			breed: normalize_opt(tag.breed),
			behaviour: normalize_opt(tag.behaviour),
			description: normalize_opt(tag.description),
			chassis_type: normalize_opt(tag.chassis_type),
			chassis_id: normalize_opt(tag.chassis_id),
			chassis_color: normalize_opt(tag.chassis_color),
			bound_to_id: &uuid!("00000000-0000-0000-0000-000000000000"), // TODO: remove once accounts exist
		},
	)
	.await?;

	Ok(Redirect::to(uri!(public::show_profile(Some(tag.ident), microchip_id))).into())
}
